#ifndef __DICTIONARY__
#define __DICTIONARY__


#include "detok.h"



/*
System-defined FCode numbers:	0x000 ... 0x7FF
Historical FCode numbers:		0x000 ... 0x2FF - subset of these fcodes
Defined FCode numbers:			0x000 ... 0x5FF - subset of these fcodes defined in the standard

Reserved FCode numbers:			0x010 ... 0x5FF - subset of these fcodes not defined in the standard
Vendor-unique FCode numbers:	0x600 ... 0x7FF - reserved for vendor-specific use with built-in devices
Program-defined FCode numbers:	0x800 ... 0xFFF - private to a particular FCode program

Power Mac G5 FCode numbers:		0x1000 ... ????	- The Power Mac G5 Quad uses out of range tokens
								???? ... 0xFFFF	- The Power Mac G5 Quad uses 0xFFFF for user defined words
*/



typedef struct token {
	s8  *name;
	u16 fcode;
	u16 flags;
	long execution_pos;
	long offset_pos;
	struct token *next;
} token_t;


/* flags */
#define kFCodeDefined				0x8000

#define kFCodeTypeMask				0x000F /* possible types 0..15 */
	#define kFCodeHistorical		0
	#define kFCodeNew				1
	#define kFCodeReserved			3
	#define kFCodeVendorUnique		4
	#define kFCodeProgramDefined	5
	#define kFCodeExtendedRange		6
	#define kFCodeNegativeRange		7

	#define kIgnoreToken			0x10
	#define kFCodePredefined		0x20

/* globals */
extern char* fcerror;
extern char* unnamed;

/* prototypes */
token_t *find_token(u16 number);
token_t *find_token_by_pos(long pos, long* offset);
token_t *find_token_by_offsetpos(long pos, long* offset);
char *get_token_name(token_t *theToken);
char *lookup_token(u16 number);
token_t *add_token(u16 number, char *name);
void init_dictionary(void);
void dump_defined();
void clear_program_tokens();

#endif
